# -*- coding: utf-8 -*- 
##############################################################################
#
# Copyright (c) Camptocamp SA - http://www.camptocamp.com
# Author: Joel Grand-Guillaume, Gille Cherix
# 
#    This file is part of the c2c_budget module
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from osv import osv, fields
from tools import SMLog
from tools import browse2dict

from wizard.smutils import *

import subprocess

# FIXME should not be needed when shortname will be deductible        
import os

class smmanager_instance(osv.osv):
    # FIXME: add validation: ports must be unique
    def _check_recursion(self, cr, uid, ids):
        level = 100
        while len(ids):
            cr.execute('select distinct parent_id from smmanager_instance where id in ('+','.join(map(str,ids))+')')
            ids = filter(None, map(lambda x:x[0], cr.fetchall()))
            if not level:
                return False
            level -= 1
        return True
    
    def _highest_assigned_port(self, cr, uid):
        # FIXME devrait prendre le highest pour le serveur de l'instance seulement
        # probleme : on ne connait pas le serveur au moment ou defaults est appele.
        cr.execute('SELECT MAX(openerp_xml_port), MAX(openerp_net_port) FROM smmanager_instance')
        highest_ports = map(lambda x: x[0], cr.fetchall())
        return max(highest_ports) or 8000

    _name = 'smmanager.instance'
    _description = 'SM Manager Instance'
    #_inherits = {'smmanager.root_branch': 'root_branch_id'}
    _columns = {
        'name' : fields.char('Name',size=30, required=True),
        'type' : fields.char('Type',size=30,required=True),
        #'path' : fields.char('Path',size=30,required=True),
        'user' : fields.char('User System',size=30),
        
        'default_path_archive' : fields.char('Default Path Archive',size=75),
        
        'openerp_admin_user' : fields.char('Openerp User Admin',size=30),
        'openerp_admin_password' : fields.char('Openerp User Password',size=30),
        
        'openerp_xml_port' : fields.integer('Port XML Openerp'),
        'openerp_net_port' : fields.integer('Port Net Openerp'),
        'webclient_port' : fields.integer('Port HTTP WebClient'),
        
        'database_user' : fields.char('Database User', size=30),
        'database_name' : fields.char('Database Name',size=30),
        'database_version' : fields.char('Database (server) Version',size=5),
        'database_password' : fields.char('Database password',size=256),
        'database_port' : fields.integer('Database Port'),
        
        'root_branch_id' : fields.many2one('smmanager.root_branch','Root branch',required=True),
        
        'parent_id' : fields.many2one('smmanager.instance','Parent Instance', required=False),
        
        'openerp_server_id': fields.many2one('smmanager.server', 'OpenERP server', required=True),
        'database_server_id': fields.many2one('smmanager.server', 'Database server', required=True),
        'web_server_id': fields.many2one('smmanager.server', 'Web server', required=True),
        
        'state': fields.selection([
			('draft','Draft'),
			('running','Running'),
			('stopped','Stopped'),
			('error','Exception')
			], 'Instance State', readonly=True, select=True),
        }
        
    _defaults = {
        'state': lambda *s: 'draft',
        'database_password': lambda *s: subprocess.Popen(["pwgen", "8", "1"], stdout=subprocess.PIPE).communicate()[0].rstrip(),
        'openerp_admin_user': lambda *s: 'admin',
        'openerp_admin_password': lambda *s: subprocess.Popen(["pwgen", "8", "1"], stdout=subprocess.PIPE).communicate()[0].rstrip(),
         
        'openerp_xml_port': lambda self, cr, uid, context: (((self._highest_assigned_port(cr, uid)/10)+1)*10)+8,
        'openerp_net_port': lambda self, cr, uid, context: (((self._highest_assigned_port(cr, uid)/10)+1)*10)+9,
        'webclient_port': lambda self, cr, uid, context: (((self._highest_assigned_port(cr, uid)/10)+1)*10)+10,

    #    'database_server_id': lambda self, cr, uid, context: context.get('id', False) and self.pool.get('smmanager.instance').browse(cr, uid, context['id']).openerp_server_id,
    #    'web_server_id': lambda self, cr, uid, context: context.get('id', False) and self.pool.get('smmanager.instance').browse(cr, uid, context['id']).openerp_server_id,
    }
    
    def _check_name(self, cr, uid, ids):
        valid = True
        for instance in self.browse(cr, uid, ids):
            # TODO : ameliorer cette condition
            # FIXME: la condition ne marche pas but : que name puisse est utilise comme nom de dossier unix
            valid &= (instance.name.find(' ') == -1)            
        return valid
        
    _constraints = [
        (_check_recursion, 'Error ! You can not create recursive instances.', ['parent_id']),
        (_check_name, 'Error ! Instance name should not contain space or special characters.', [])
    ]

    def write(self, cr, uid, ids, vals, context={}):
        parent_return_value = osv.osv.write(self, cr, uid, ids, vals, context)
        print "salut!"
        return parent_return_value

    def start(self, cr, uid, ids, vals, context={}):
        SMLog.SMLog(cr, uid, "Starting...", instance_id=ids[0])

        instance_obj = self.pool.get('smmanager.instance')
        instance = instance_obj.browse(cr, uid, ids[0])

        connection = instance.openerp_server_id.connect()

        try:
            remoteAdmin(connection, "startInstance('%s')" % instance.name)
            instance_obj.write(cr, uid, ids[0], {'state': 'running' })
            SMLog.SMLog(cr, uid, "Started instance.", instance_id=ids[0])
        except Exception, e:
            SMLog.SMLog(cr, uid, "Could not start instance because of %s" % e, instance_id=ids[0])
            instance_obj.write(cr, uid, ids[0], {'state': 'error' })

        instance.openerp_server_id.disconnect(connection)

        return {}
        
    def stop(self, cr, uid, ids, vals, context={}):
        SMLog.SMLog(cr, uid, "Stopping...", instance_id=ids[0])
    
        instance_obj = self.pool.get('smmanager.instance')
        instance = instance_obj.browse(cr, uid, ids[0])
    
        connection = instance.openerp_server_id.connect()
    
        try:
            remoteAdmin(connection, "stopInstance('%s')" % instance.name)
            instance_obj.write(cr, uid, ids[0], {'state': 'stopped' })
            SMLog.SMLog(cr, uid, "Stopped instance.", instance_id=ids[0])
        except Exception, e:
            SMLog.SMLog(cr, uid, "Could not stop instance because of %s" % e, instance_id=ids[0])
            instance_obj.write(cr, uid, ids[0], {'state': 'error' })
    
        instance.openerp_server_id.disconnect(connection)
    
        return {}
            
    def restart(self, cr, uid, ids, vals, context={}):
        SMLog.SMLog(cr, uid, "Restarted...", instance_id=ids[0])
        
        instance_obj = self.pool.get('smmanager.instance')
        instance = instance_obj.browse(cr, uid, ids[0])
        
        connection = instance.openerp_server_id.connect()
        
        #if instance.state == 'running':
        try:
            remoteAdmin(connection, "stopInstance('%s')" % instance.name)
            instance_obj.write(cr, uid, ids[0], {'state': 'stopped' })
            SMLog.SMLog(cr, uid, "Stopped instance.", instance_id=ids[0])
        except Exception, e:
            SMLog.SMLog(cr, uid, "Could not stop instance because : %s" % e, instance_id=ids[0])
            instance_obj.write(cr, uid, ids[0], {'state': 'error' })

        #if instance.state == 'stopped':
        try:
            remoteAdmin(connection, "startInstance('%s')" % instance.name)
            instance_obj.write(cr, uid, ids[0], {'state': 'running' })
            SMLog.SMLog(cr, uid, "Started instance.", instance_id=ids[0])
        except Exception, e:
            SMLog.SMLog(cr, uid, "Could not start instance because : %s" % e, instance_id=ids[0])
            instance_obj.write(cr, uid, ids[0], {'state': 'error' })
        
        instance.openerp_server_id.disconnect(connection)
        
        return {}
    
    #
    # create a remote archive of the instance
    #
    def archive(self, cr, uid, ids, vals, context={}):
        SMLog.SMLog(cr, uid, "Archiving...", instance_id=ids[0])
        instance = self.browse(cr, uid, ids[0])
        
        archive_name = vals['form']['archive_name']
        instance_dict = browse2dict(instance)
        instance_dict['archive_name'] = archive_name
        
        #
        # FIXME: should ensure connection with server (using server.state)
        #
        try:
            import pdb; pdb.set_trace()
            remote_result = remoteAdmin(vals['connection'], "dumpInstance(%s, database_only=%s)" % (instance_dict, vals['form']['database_only']))
            
            if len(remote_result) == 0 or remote_result[0] == '':
                raise "Nothing to archive!"
            
            SMLog.SMLog(cr, uid, "Archive : %s" % remote_result, instance_id=ids[0])
            
            archive_obj = self.pool.get('smmanager.archive')
            archive_id = archive_obj.create(cr, uid, {'description': vals['form']['description'], 'path': archive_name, 'server_id': instance.openerp_server_id.id, 'instance_id': ids[0] })
            archive = archive_obj.browse(cr, uid, archive_id)
        except Exception, e:
            SMLog.SMLog(cr, uid, "Could not archive because : %s" % e, instance_id=ids[0])

        return {}

    # FIXME: pass dict with server info in order to support separated db and webclient servers
    def deploy(self, cr, uid, ids, vals, context={}):
        SMLog.SMLog(cr, uid, "Trying to deploy.", instance_id=ids[0])
        instance = self.browse(cr, uid, ids[0])
        
        # FIXME: just used to browse to branch_ids so that the ORM load'em... REFACTOR!
        test_dict = browse2dict(instance.root_branch_id, with_relations=True)
        branches = dict(map(lambda b: (b.name, b.path), instance.root_branch_id.branch_ids))

        instance_dict = browse2dict(instance, with_relations=True)
        
        #
        # FIXME: should ensure connection with server (using server.state or server.connection when hops will be theres)
        #
        result = ""
        try:
            if vals['form']['archive_id']:
                SMLog.SMLog(cr, uid, "Deploy : deploying from file.", instance_id=ids[0])
                
                # send archive
                archive_obj = self.pool.get('smmanager.archive')
                archive = archive_obj.browse(cr, uid, vals['form']['archive_id'])
                
                source_server = archive.server_id
                destination_server = instance.openerp_server_id
                
                if instance.openerp_server_id == archive.server_id:
                    # first transfer the file
                    raise "NotImplemented"
                    
                
                import pdb; pdb.set_trace()
                # send file to server
                # archive.path to instance.openerp_server_id
                #ssh = archive.openerp_server_id._ssh_connection(cr, uid, [archive.openerp_server_id.id], context=vals)
                #
                #ftp = ssh.open_sftp()
                #ftp.get(archive.path, archive_name)
                #ftp.close()
                #
                #ssh.close()
                archive_dict = browse2dict(archive)
                result = remoteAdmin(vals['connection'], "deployInstance(%s, %s)" % (instance_dict, archive_dict))
                SMLog.SMLog(cr, uid, "Deploy : %s." % result, instance_id=ids[0])
            else:
                SMLog.SMLog(cr, uid, "Deploy : creating a new instance from %s." % instance.root_branch_id.name, instance_id=ids[0])

                results = remoteAdmin(vals['connection'], "createInstance(%s)" % instance_dict)
                
                map(lambda result: SMLog.SMLog(cr, uid, "Deploy : %s." % result, instance_id=ids[0]), results)

            self.write(cr, uid, ids[0], {'state': 'stopped' })
        except Exception, e:
            SMLog.SMLog(cr, uid, "Could not deploy because : %s." % e, instance_id=ids[0])
            self.write(cr, uid, ids[0], {'state': 'error' })
        
        return result

    #
    # Create a tunnel chain to the OpenERP instance
    #
    def connect(self, cr, uid, ids, context = {}):
        instance = self.browse(cr, uid, ids[0])
        
        connection = instance.openerp_server_id.connect({ 'destination_port' : instance.port_net_openerp })

        connection['openerp_admin_user'] = instance.openerp_admin_user
        connection['openerp_admin_password'] = instance.openerp_admin_password
        
        return connection
    
    #
    # Create a tunnel chain to the OpenERP instance
    #
    def connect_database(self, cr, uid, ids, context = {}):
        instance = self.browse(cr, uid, ids[0])
    
        connection = instance.database_server_id.connect({ 'destination_port' : instance.database_port })
        
        connection['database_name'] = instance.database_name
        connection['database_user'] = instance.database_user
        connection['database_password'] = instance.database_password
  
        return connection
smmanager_instance()