# -*- coding: utf-8 -*- 
##############################################################################
#
# Copyright (c) Camptocamp SA - http://www.camptocamp.com
# Author: Joel Grand-Guillaume, Gille Cherix
# 
#    This file is part of the c2c_budget module
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from osv import osv, fields
from wizard.smutils import *

import os
import time
import subprocess

from tools import SMLog

class smmanager_server_type (osv.osv):
    _name = 'smmanager.server.type'
    _description = 'SM Manager Server type'
    _columns = {
        'name' : fields.char("Type name", size=60, required=True, translate=True),
        'comment' : fields.text("Note"),
    }
smmanager_server_type()

class smmanager_server(osv.osv):
    _name = 'smmanager.server'
    _description = 'SM Manager Server'
    _columns = {
        'name' : fields.char('Server Name', size=30, required=True),
        
        'host_name' : fields.char('Server Host Name',size=30,required=True),
        
        'server_type_ids': fields.many2many(   
                                    'smmanager.server.type',
                                    'smmanager_server_type_rel',
                                    'type_id',
                                    'server_id',
                                    'Server types'
                                ),
        
        'ssh_user':fields.char('SSH user', size=256, required=True),
        'ssh_port':fields.integer('SSH Port', required=True),
        'ssh_password':fields.char('SSH Password', size=256),
        'ssh_args':fields.char('SSH args', size=512),
        
        'connection_ids': fields.one2many('smmanager.connection', 'destination_server_id', 'Gateways'),
        
        'instances_home' : fields.char('Instance home',size=30),

        # TODO: add connection state
        #'state': fields.selection([
		#	('up','Up'),
		#	('down','Down'),
		#	('connected','Connected'),
		#	('error','Exception')
		#	], 'Server State', readonly=True, select=True),
        }
        
    #
    # create a paramiko ssh connection
    #
    def _ssh_connection(self, cr, uid, ids, context = {}):
        #
        # only used for scp at the moment.
        # TODO: corriger l'abstraction (quand le serveur aura un connection state)
        if context['connection']:
            # connection via un tunnel
            server = context['connection']
        else:
            # connection directe
            server = self.browse(cr, uid, ids[0])
        
        import paramiko

        ssh = paramiko.SSHClient()
        ssh.load_system_host_keys()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        ssh.connect(server['host_name'], port=int(server['port']), username=server['ssh_user'], password=server['ssh_password'])
        return ssh

    #
    # Create a tunnel chain to the server
    #
    def connect(self, cr, uid, ids, context = {}):
        log_obj = self.pool.get('smmanager.log')
        
        if context.has_key('destination_port'):
            destination_port = context['destination_port']
            log = log_obj.create(cr, uid, { 'description' : "Called server connect with destination port %s." % destination_port })
        else:
            log = log_obj.create(cr, uid, { 'description' : "Called server connect without destination port." })
        
        server = self.browse(cr, uid, ids[0])

        gateways = map(lambda connection_id: connection_id.server_id, server.connection_ids)
        gateways.append(server)

        if (len(gateways) == 0):
            return { 'host_name': server.host_name, 'ssh_user': server.ssh_user, 'ssh_password': server.ssh_password, 'port': server.ssh_port }

        #
        # for each hop to the server, create a tunnel
        #
        ports = free_ports(len(gateways)+1)

        # generate ssh command argument for the tunnels creation
        tunnels = []
        if len(gateways) == 1:
            tunnels.append(ssh_tunnel(gateways[0].host_name, gateways[0].ssh_user, gateways[0].ssh_port, ports[0], 'localhost', gateways[0].ssh_port, remote_connection=True))
        else:
            tunnels.append(ssh_tunnel(gateways[0].host_name, gateways[0].ssh_user, gateways[0].ssh_port, ports[0], 'localhost', gateways[0].ssh_port))

            for i in range(1, len(gateways)):
                h = gateways[i]
                if i == len(gateways)-1:
                    if not context.has_key('destination_port'):
                        destination_port = h.ssh_port
                        tunnels.append(ssh_tunnel('localhost', gateways[i-1].ssh_user, ports[i-1], ports[i], h.host_name, destination_port, remote_connection=True))
                    else:
                        tunnels.append(ssh_tunnel('localhost', gateways[i-1].ssh_user, ports[i-1], ports[i], h.host_name, h.ssh_port))

        # launch ssh command waiting 1 sec between each for connection establishment
        # TODO: Gestion d'exception
        jobs = []
        try:
            jobs = map(lambda t: time.sleep(1) or subprocess.Popen(t, stdout=subprocess.PIPE, stderr=subprocess.PIPE), tunnels)
            pids = map(lambda p: p.pid, jobs)
            
            #FIXME replace with a proper check that the connection has been established
            time.sleep(3)
            
            # FIXME: bloque le serveur openerp, mais est necessaire pour les logs!
            #map(lambda p: SMLog.SMLog(cr, uid, p.communicate()[0] + p.communicate()[1], server_id=ids[0]), jobs)
            SMLog.SMLog(cr, uid, "Tunnel connected.", server_id=ids[0])
        except Exception, e:
            SMLog.SMLog(cr, uid, e, server_id=ids[0])
						raise "Problem with tunnel connection, check the connections hops."

        # FIXME: won't work when /etc/hosts point the local name to 127.0.0.1
        # FIXME: should be done in only one place
        local_host_name = socket.gethostbyaddr(socket.gethostname())[2][0]
        
        local_port = ports[len(tunnels)-1]
        
        return {'port':         str(local_port), # FIXME: str ici et int() dans _ssh_connection. Stupide?
                'host_name':    local_host_name,
                'ssh_user':     server.ssh_user,
                'ssh_password': server.ssh_password,
                'ssh_pids':     pids, # ','.join(map(lambda p: str(p), pids)) or ''
                'command':      "ssh %s@%s -p %i" % (server.ssh_user, local_host_name, local_port) }

    # REFACTOR: not used at the moment
    def disconnect(self, cr, uid, ids, context = {}):
        log_obj = self.pool.get('smmanager.log')
        log = log_obj.create(cr, uid, { 'description' : "Called server disconnect." })
        
        # FIXME: gestion d'erreur
        try:
            map(lambda pid: time.sleep(2) or os.kill(pid, 3), context['ssh_pids'])
        except Exception, e:
            SMLog.SMLog(cr, uid, e, server_id=ids[0])
        
        return {}
        
smmanager_server()