#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
# -*- mode: python; indent-tabs-mode: nil; tab-width: 4 -*-
# vim: set tabstop=2 shiftwidth=2 expandtab:
# -*- coding: utf-8 -*- 
##############################################################################
#
# Copyright (c) Camptocamp SA - http://www.camptocamp.com
# Author: Joel Grand-Guillaume, Gille Cherix
# 
#    This file is part of the c2c_budget module
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
###############################################################################

__version__ = '0.1.1'
__author__    = 'Mathieu Bornoz', 'mathieu.bornoz@camptocamp.com'
__date__        = "2008-12-10"

import os
import sys
import glob
import getopt
import ConfigParser
from string import Template
from commands import getstatusoutput

# FIXME: should be passed as argument
#_INSTANCE_ROOT="/home/cherix/instances"

_debug = open('/tmp/remote_admin-debug-%d' % os.getpid()  , 'w')

def error(msg):
    if _debug:
        _debug.write("[error] %s\n" % msg)
        _debug.flush()
    #channel.send("[error] %s" % msg)
    channel.close() # FIXME: vérifier que ça interrompt bien la connexion (et donc le code en cours d'exécution)
    
def execCmd(cmd, msg):
    status, output = getstatusoutput(cmd)
    
    if status:
        if _debug:
            _debug.write("[error] %s failed !\n" % msg)
            _debug.flush()
        channel.send("[error] %s failed !" % msg)
    else:
        if _debug:
            _debug.write("[output] : %s\n" % msg)
            _debug.flush()
        channel.send("[output] : %s" % output)
    return output

def listDb(instance):
    channel.send(os.system("psql -l -h %s" % server['host_name']))

def _database_exists(name):
    return not os.system("psql -l -h %s |grep -wq %s" % (server['host_name'], name))

def dumpDb(instance, name):
    dbname = "%s" % name
    path = os.path.join(server['instances_home'], name)
    
    execCmd('pg_dump %s > %s/database/dump.sql' % (dbname, path), "dumping database") # FIXME: nom variable en fct de date

def loadDb(instance, name):
    dbname = "%s" % name
    path = os.path.join(server['instances_home'], name)
    
    execCmd('/usr/bin/psql %s -U %s -h %s < %s/database/dump.sql' % (dbname, name, server['host_name'], path), "reload database from dump")

def dumpInstance(instance, database_only=False):
    name = instance['name']
    archive_name = instance['archive_name']

    # dump db
    if (_database_exists(name)):
        dumpDb(name)
    else:
        error('no database named %s' % name)

    # create tarball and transfer
    channel.send("[output] creating archive...")
    if database_only:
        execCmd('cd %s; tar czf %s %s/database' % (server['instances_home'], archive_name, name), "archiving")
    else:
        execCmd('cd %s; tar czf %s %s' % (server['instances_home'], archive_name, name), "archiving")

def deployInstance(instance, archive):
    raise "NOT IMPLEMENTED"
    
    # extract os.path.join(server['instances_home'], archive.path)
    execCmd('cd %s; tar xzf %s' % (server['instances_home'], archive_name, name), "archiving")
    
    createUserDb(instance['name'], instance['database_password'])
    createDb(instance['name'])
    loadDb(instance['name'])
    
    writeConfig(instance['name'], instance['openerp_xml_port'], instance['openerp_net_port'], instance['database_password'])

    

def createDb(name):
    dbname = "%s" % name

    if _database_exists(name):
        error("%s database already exists" % dbname)
    else:
        execCmd("createdb -h %s --encoding utf8 --owner %s %s" % (server['host_name'], dbname, dbname), "create postgres database %s" % dbname)

def createUserDb(name, password):
    dbuser = "%s" % name

    output = execCmd("psql -h %s -c '\du %s' template1" % (server['host_name'], dbuser), "check if %s user exists" % dbuser)
        
    if "(1 " in output:
        error("%s user already exists" % dbuser)

    if "(0 " in output:
        execCmd("createuser -h %s -S -D -l -R %s" % dbuser, "create postgres user %s" % (server['host_name'], dbuser))
        execCmd("psql -h %s -c \"ALTER USER %s WITH PASSWORD '%s';\" template1"    % (server['host_name'], dbuser, password), "set password for user %s" % dbuser)
        return password

def createInstance(instance):
    import pdb; pdb.set_trace()
    createInstanceTree(instance['name'], instance['root_branch_id']['branch_ids'])
    
    createUserDb(instance['name'], instance['database_password'])
    createDb(instance['name'])
    
    writeConfig(instance['name'], instance['openerp_xml_port'], instance['openerp_net_port'], instance['database_password'])
    
def createInstanceTree(name, branches):
    if not branches:
        error('no argument branches was set')
    
    instDir = os.path.join(server['instances_home'], name)
    
    if os.path.isdir(instDir):
        error("an instance already exists with name %s" % name)
    
    os.mkdir(instDir)
    os.mkdir(os.path.join(instDir, "log"))
    os.mkdir(os.path.join(instDir, "src"))
    os.mkdir(os.path.join(instDir, "run"))
    os.mkdir(os.path.join(instDir, "database"))
    
    for branch in branches:
        execCmd("bzr co %s %s" % (branch['path'], os.path.join(instDir, "src", branch['name'])), "checkout of the code")

def updateInstanceTree(instance):
    for branch in instance['root_branch_id']['branch_ids']:
        execCmd("bzr merge %s %s" % (branch['path'], os.path.join(instDir, "src", branch['name'])), "update of the code")

def deleteInstance(name):
    #channel.send("dropdb %s" % name)
    #channel.send("dropuser %s" % name)
    #channel.send("rm -fR %s/%s" % (server['instances_home'], name))

    execCmd("dropdb %s" % name, "removing db")
    execCmd("dropuser %s" % name, "remove db user")
    execCmd("rm -fR %s/%s" % (server['instances_home'], name), "removing code")

def getLogFilePath(name):
    return "%s/%s/log/server.log" % (server['instances_home'], name)

def getConfigPath(name):
    return "%s/%s/instance.ini" % (server['instances_home'], name)

def getPidFilePath(name):
    return "%s/%s/run/server.pid" % (server['instances_home'], name)

def getServerScript(name):
    return "%s/%s/src/server/bin/openerp-server.py" % (server['instances_home'], name)

def writeConfig(name, port, net_port, db_password):
    config = ConfigParser.RawConfigParser()
    config.add_section('Main')
    config.set('Main', 'options', ' --without-demo=all')
    config.set('Main', 'logfile', getLogFilePath(name))
    config.set('Main', 'db_host', 'localhost')
    config.set('Main', 'db_port', '5432')
    config.set('Main', 'db_password', db_password)
    config.set('Main', 'db_user', "%s" % name)
    config.set('Main', 'database', "%s" % name)
    config.set('Main', 'smtp', '')
    config.set('Main', 'net_port', net_port)
    config.set('Main', 'port', port)
    config.set('Main', 'startonboot', 0)

    f = open(getConfigPath(name), 'wb')
    config.write(f)

def getConfig(name):
    print 'getConfig'
    config = ConfigParser.ConfigParser()
    config.read(getConfigPath(name))
    return config

def getConfigOption(name, section, option):
    cfg = getConfig(name)
    return cfg.get(section, option)

def getDictConfig(name, section):
    tmpDict = {}
    conf = getConfig(name)
    items = conf.items(section)
    for i in items:
        tmpDict[i[0]] = i[1]
    return tmpDict

def getNextTwoFreePort():
    portList = []
    config = ConfigParser.ConfigParser()
    config.read("%s/instances.ini" % server['instances_home'])
    try :   
        portList.append(config.get('Main', 'port'))
        portList.append(config.get('Main', 'net_port'))
    except :
        config.add_section('Main')
        config.set('Main', 'net_port', 8070)
        config.set('Main', 'port', 8069)
        f = open("%s/instances.ini" % server['instances_home'], 'wb')
        config.write(f)
        

    if portList == []:
        return 8069, 8070 
    else:
        portList.sort()
        return int(portList[-1])+1, int(portList[-1])+2    

def startInstance(name):
    cfg = getDictConfig(name, 'Main')
    print cfg

    cmdDaemon = "/sbin/start-stop-daemon --start --pidfile %s --background --make-pidfile --exec %s  -- %s"
    cmdDaemonRoot = "/sbin/start-stop-daemon --start --pidfile %s --chuid openerp --background --make-pidfile --exec %s  -- %s"

    openerpArgs = "%(options)s --port=%(port)s "

    if cfg['db_host']:
        openerpArgs += "--db_host=%(db_host)s "

    if cfg['db_password']:
        openerpArgs += "--db_password=%(db_password)s "

    if cfg['db_port']:
        openerpArgs += "--db_port=%(db_port)s "

    openerpArgs += "--logfile=%(logfile)s "
    openerpArgs += "--db_user=%(db_user)s "
    openerpArgs += "--net_port=%(net_port)s --smtp=%(smtp)s"


    if os.getuid() == 0 :
        print '--------------------UID 0----------------------'
        print cmdDaemonRoot %(getPidFilePath(name), getServerScript(name), openerpArgs % cfg)
        os.system(cmdDaemonRoot %(getPidFilePath(name), getServerScript(name), openerpArgs % cfg))

    else:
        print '---------------UID other--------------'
        print cmdDaemon %(getPidFilePath(name), getServerScript(name), openerpArgs % cfg)
        os.system(cmdDaemon %(getPidFilePath(name), getServerScript(name), openerpArgs % cfg)),  
        print '------------------------'
    os.system("chown openerp: %s" % getPidFilePath(name))

def stopInstance(name):
    execCmd("/sbin/start-stop-daemon -q --stop --pidfile %s --oknodo" % getPidFilePath(name), "stop instance")
    if os.path.exists(getPidFilePath(name)):
        os.remove(getPidFilePath(name))
    # have to check if there's another fork... strange behaviour we detected.
    output = execCmd('''ps fax | awk '$6 ~ /%s.+openerp/ {print $1}' ''' % name, 'get pid')
    if output:
      execCmd('kill -6 %s'%output, 'really kill %s'%name)

def listInstance():
    print " ", 
    print "NAME".ljust(12), 
    print "PID".ljust(12), 
    print "STATUS".ljust(12), 
    print "BOOTSTART".ljust(12),
    print "PORT".ljust(12),
    print "NETPORT".ljust(12)

    instances = glob.glob("%s/*" % server['instances_home'])
    instances.sort()

    for instance in instances:
        name = os.path.basename(instance)

        pid = execCmd("pgrep -f %s" % instance, "get pid")
        if "\n" in pid:
            pid = int(pid.split('\n')[0])
        else: pid = 0
     
        pidfile = -1
        if os.path.isfile(getPidFilePath(name)):
            pidfile = int(open(getPidFilePath(name)).read())

        if pid == pidfile:
            print " ", 
            print name.ljust(12), 
            print str(pid).ljust(12), 
            print "running".ljust(12), 
            print getConfigOption(name, 'Main', 'startonboot').ljust(12),
            print getConfigOption(name, 'Main', 'port').ljust(12),
            print getConfigOption(name, 'Main', 'net_port').ljust(12)
        else:
            print " ", 
            print name.ljust(12), 
            print "---".ljust(12), 
            print "stopped".ljust(12), 
            print getConfigOption(name, 'Main', 'startonboot').ljust(12),
            print getConfigOption(name, 'Main', 'port').ljust(12),
            print getConfigOption(name, 'Main', 'net_port').ljust(12)


def version(name):
    print "%s v%s"%(name,__version__)
    print "Write by %s, %s\n"%(__author__[0],__author__[1])


if __name__=="__channelexec__":
    for fn in channel:
        channel.send(eval(fn))
